@@ -1,6 +1,9 @@
 from contextlib import contextmanager
 from enum import Enum
 import os
+import pytest  # type: ignore
+import subprocess
+from subprocess import PIPE
 import sys
 import tempfile
 from typing import Tuple, List, Generator, Optional
@@ -9,7 +12,6 @@ from unittest import TestCase, main
 import mypy.api
 from mypy.modulefinder import get_site_packages_dirs
 from mypy.test.config import package_path
-from mypy.test.helpers import run_command
 from mypy.util import try_find_python2_interpreter
 
 # NOTE: options.use_builtins_fixtures should not be set in these
@@ -52,8 +54,8 @@ import typedpkg_ns.ns.bbb; bf = typedpkg_ns.ns.bbb.bf"""
 
 class SimpleMsg(Enum):
     msg_dne = "{tempfile}:3: error: Module 'typedpkg' has no attribute 'dne'"
-    msg_list = "{tempfile}:5: error: Revealed type is 'builtins.list[builtins.str]'"
-    msg_tuple = "{tempfile}:5: error: Revealed type is 'builtins.tuple[builtins.str]'"
+    msg_list = "{tempfile}:5: note: Revealed type is 'builtins.list[builtins.str]'"
+    msg_tuple = "{tempfile}:5: note: Revealed type is 'builtins.tuple[builtins.str]'"
 
 
 class NamespaceMsg(Enum):
@@ -112,6 +114,7 @@ class ExampleProg(object):
             old_dir = os.getcwd()
             os.chdir(venv_dir)
         try:
+            cmd_line.append('--no-error-summary')
             if python_executable != sys.executable:
                 cmd_line.append('--python-executable={}'.format(python_executable))
             out, err, returncode = mypy.api.run(cmd_line)
@@ -135,13 +138,13 @@ class TestPEP561(TestCase):
         # Sadly, we need virtualenv, as the Python 3 venv module does not support creating a venv
         # for Python 2, and Python 2 does not have its own venv.
         with tempfile.TemporaryDirectory() as venv_dir:
-            returncode, lines = run_command([sys.executable,
-                                             '-m',
-                                             'virtualenv',
-                                             '-p{}'.format(python_executable),
-                                            venv_dir], cwd=os.getcwd())
-            if returncode != 0:
-                err = '\n'.join(lines)
+            proc = subprocess.run([sys.executable,
+                                   '-m',
+                                   'virtualenv',
+                                   '-p{}'.format(python_executable),
+                                   venv_dir], cwd=os.getcwd(), stdout=PIPE, stderr=PIPE)
+            if proc.returncode != 0:
+                err = proc.stdout.decode('utf-8') + proc.stderr.decode('utf-8')
                 self.fail("Failed to create venv. Do you have virtualenv installed?\n" + err)
             if sys.platform == 'win32':
                 yield venv_dir, os.path.abspath(os.path.join(venv_dir, 'Scripts', 'python'))
@@ -165,9 +168,9 @@ class TestPEP561(TestCase):
                 install_cmd.append('develop')
             else:
                 install_cmd.append('install')
-        returncode, lines = run_command(install_cmd, cwd=working_dir)
-        if returncode != 0:
-            self.fail('\n'.join(lines))
+        proc = subprocess.run(install_cmd, cwd=working_dir, stdout=PIPE, stderr=PIPE)
+        if proc.returncode != 0:
+            self.fail(proc.stdout.decode('utf-8') + proc.stderr.decode('utf-8'))
 
     def setUp(self) -> None:
         self.simple_prog = ExampleProg(SIMPLE_PROGRAM)
@@ -186,6 +189,9 @@ class TestPEP561(TestCase):
         dirs = get_site_packages_dirs(sys.executable)
         assert dirs
 
+    @pytest.mark.skipif(sys.platform == 'darwin' and hasattr(sys, 'base_prefix') and
+                        sys.base_prefix != sys.prefix,
+                        reason="Temporarily skip to avoid having a virtualenv within a venv.")
     def test_typedpkg_stub_package(self) -> None:
         self.simple_prog.create()
         with self.virtualenv() as venv:
@@ -197,6 +203,9 @@ class TestPEP561(TestCase):
                 venv_dir=venv_dir,
             )
 
+    @pytest.mark.skipif(sys.platform == 'darwin' and hasattr(sys, 'base_prefix') and
+                        sys.base_prefix != sys.prefix,
+                        reason="Temporarily skip to avoid having a virtualenv within a venv.")
     def test_typedpkg(self) -> None:
         self.simple_prog.create()
         with self.virtualenv() as venv:
@@ -208,6 +217,40 @@ class TestPEP561(TestCase):
                 venv_dir=venv_dir,
             )
 
+    def test_mypy_path_is_respected(self) -> None:
+        packages = 'packages'
+        pkg_name = 'a'
+        with tempfile.TemporaryDirectory() as temp_dir:
+            old_dir = os.getcwd()
+            os.chdir(temp_dir)
+            try:
+                # Create the pkg for files to go into
+                full_pkg_name = os.path.join(temp_dir, packages, pkg_name)
+                os.makedirs(full_pkg_name)
+
+                # Create the empty __init__ file to declare a package
+                pkg_init_name = os.path.join(temp_dir, packages, pkg_name, '__init__.py')
+                open(pkg_init_name, 'w', encoding='utf8').close()
+
+                mypy_config_path = os.path.join(temp_dir, 'mypy.ini')
+                with open(mypy_config_path, 'w') as mypy_file:
+                    mypy_file.write('[mypy]\n')
+                    mypy_file.write('mypy_path = ./{}\n'.format(packages))
+
+                with self.virtualenv() as venv:
+                    venv_dir, python_executable = venv
+
+                    cmd_line_args = []
+                    if python_executable != sys.executable:
+                        cmd_line_args.append('--python-executable={}'.format(python_executable))
+                    cmd_line_args.extend(['--config-file', mypy_config_path,
+                                          '--package', pkg_name])
+
+                    out, err, returncode = mypy.api.run(cmd_line_args)
+                    assert returncode == 0
+            finally:
+                os.chdir(old_dir)
+
     def test_stub_and_typed_pkg(self) -> None:
         self.simple_prog.create()
         with self.virtualenv() as venv:
@@ -220,6 +263,9 @@ class TestPEP561(TestCase):
                 venv_dir=venv_dir,
             )
 
+    @pytest.mark.skipif(sys.platform == 'darwin' and hasattr(sys, 'base_prefix') and
+                        sys.base_prefix != sys.prefix,
+                        reason="Temporarily skip to avoid having a virtualenv within a venv.")
     def test_typedpkg_stubs_python2(self) -> None:
         self.simple_prog.create()
         python2 = try_find_python2_interpreter()
@@ -246,6 +292,9 @@ class TestPEP561(TestCase):
                     venv_dir=venv_dir,
                 )
 
+    @pytest.mark.skipif(sys.platform == 'darwin' and hasattr(sys, 'base_prefix') and
+                        sys.base_prefix != sys.prefix,
+                        reason="Temporarily skip to avoid having a virtualenv within a venv.")
     def test_typedpkg_egg(self) -> None:
         self.simple_prog.create()
         with self.virtualenv() as venv:
@@ -257,6 +306,9 @@ class TestPEP561(TestCase):
                 venv_dir=venv_dir,
             )
 
+    @pytest.mark.skipif(sys.platform == 'darwin' and hasattr(sys, 'base_prefix') and
+                        sys.base_prefix != sys.prefix,
+                        reason="Temporarily skip to avoid having a virtualenv within a venv.")
     def test_typedpkg_editable(self) -> None:
         self.simple_prog.create()
         with self.virtualenv() as venv:
@@ -268,6 +320,9 @@ class TestPEP561(TestCase):
                 venv_dir=venv_dir,
             )
 
+    @pytest.mark.skipif(sys.platform == 'darwin' and hasattr(sys, 'base_prefix') and
+                        sys.base_prefix != sys.prefix,
+                        reason="Temporarily skip to avoid having a virtualenv within a venv.")
     def test_typedpkg_egg_editable(self) -> None:
         self.simple_prog.create()
         with self.virtualenv() as venv:
